@@ -1,8 +1,7 @@
 use crate::csv::{CsvEncoding, NullValues};
 use crate::csv_core::utils::*;
 use crate::csv_core::{buffer::*, parser::*};
-use crate::mmap::MmapBytesReader;
-use crate::utils::resolve_homedir;
+use crate::mmap::ReaderBytes;
 use crate::PhysicalIoExpr;
 use crate::ScanAggregation;
 use polars_arrow::array::*;
@@ -12,20 +11,16 @@ use rayon::prelude::*;
 use rayon::ThreadPoolBuilder;
 use std::borrow::Cow;
 use std::fmt;
-#[cfg(feature = "decompress")]
-use std::io::SeekFrom;
-use std::path::PathBuf;
 use std::sync::atomic::Ordering;
 use std::sync::{atomic::AtomicUsize, Arc};
 
 /// CSV file reader
-pub struct CoreReader<'a, R: MmapBytesReader> {
+pub(crate) struct CoreReader<'a> {
+    reader_bytes: Option<ReaderBytes<'a>>,
     /// Explicit schema for the CSV file
     schema: Cow<'a, Schema>,
     /// Optional projection for which columns to load (zero-based column indices)
     projection: Option<Vec<usize>>,
-    /// File reader
-    reader: Option<R>,
     /// Current line number, used in error reporting
     line_number: usize,
     ignore_parser_errors: bool,
@@ -33,23 +28,19 @@ pub struct CoreReader<'a, R: MmapBytesReader> {
     n_rows: Option<usize>,
     encoding: CsvEncoding,
     n_threads: Option<usize>,
-    path: Option<PathBuf>,
     has_header: bool,
     delimiter: u8,
     sample_size: usize,
     chunk_size: usize,
     low_memory: bool,
     comment_char: Option<u8>,
+    quote_char: Option<u8>,
     null_values: Option<Vec<String>>,
-    /// If schema was given by user or inferred
-    #[cfg(feature = "decompress")]
-    inferred_schema: bool,
+    predicate: Option<Arc<dyn PhysicalIoExpr>>,
+    aggregate: Option<&'a [ScanAggregation]>,
 }
 
-impl<'a, R> fmt::Debug for CoreReader<'a, R>
-where
-    R: MmapBytesReader,
-{
+impl<'a> fmt::Debug for CoreReader<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("Reader")
             .field("schema", &self.schema)
@@ -107,7 +98,126 @@ impl RunningSize {
     }
 }
 
-impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
+impl<'a> CoreReader<'a> {
+    #[allow(clippy::too_many_arguments)]
+    pub(crate) fn new(
+        reader_bytes: ReaderBytes<'a>,
+        n_rows: Option<usize>,
+        mut skip_rows: usize,
+        mut projection: Option<Vec<usize>>,
+        max_records: Option<usize>,
+        delimiter: Option<u8>,
+        has_header: bool,
+        ignore_parser_errors: bool,
+        schema: Option<&'a Schema>,
+        columns: Option<Vec<String>>,
+        encoding: CsvEncoding,
+        n_threads: Option<usize>,
+        schema_overwrite: Option<&'a Schema>,
+        dtype_overwrite: Option<&'a [DataType]>,
+        sample_size: usize,
+        chunk_size: usize,
+        low_memory: bool,
+        comment_char: Option<u8>,
+        quote_char: Option<u8>,
+        null_values: Option<NullValues>,
+        predicate: Option<Arc<dyn PhysicalIoExpr>>,
+        aggregate: Option<&'a [ScanAggregation]>,
+    ) -> Result<CoreReader<'a>> {
+        #[cfg(any(feature = "decompress", feature = "decompress-fast"))]
+        let mut reader_bytes = reader_bytes;
+
+        #[cfg(not(any(feature = "decompress", feature = "decompress-fast")))]
+        if is_compressed(&reader_bytes) {
+            return Err(PolarsError::ComputeError("cannot read compressed csv file; compile with feature 'decompress' or 'decompress-fast'".into()));
+        }
+
+        // check if schema should be inferred
+        let delimiter = delimiter.unwrap_or(b',');
+
+        let mut schema = match schema {
+            Some(schema) => Cow::Borrowed(schema),
+            None => {
+                #[cfg(any(feature = "decompress", feature = "decompress-fast"))]
+                {
+                    // We keep track of the inferred schema bool
+                    // In case the file is compressed this schema inference is wrong and has to be done
+                    // again after decompression.
+                    if let Some(b) = decompress(&reader_bytes) {
+                        reader_bytes = ReaderBytes::Owned(b);
+                    }
+
+                    let (inferred_schema, _) = infer_file_schema(
+                        &reader_bytes,
+                        delimiter,
+                        max_records,
+                        has_header,
+                        schema_overwrite,
+                        &mut skip_rows,
+                        comment_char,
+                        quote_char,
+                    )?;
+                    Cow::Owned(inferred_schema)
+                }
+                #[cfg(not(any(feature = "decompress", feature = "decompress-fast")))]
+                {
+                    let (inferred_schema, _) = infer_file_schema(
+                        &reader_bytes,
+                        delimiter,
+                        max_records,
+                        has_header,
+                        schema_overwrite,
+                        &mut skip_rows,
+                        comment_char,
+                        quote_char,
+                    )?;
+                    Cow::Owned(inferred_schema)
+                }
+            }
+        };
+        if let Some(dtypes) = dtype_overwrite {
+            let mut s = schema.into_owned();
+            let fields = s.fields_mut();
+            for (dt, field) in dtypes.iter().zip(fields) {
+                *field = Field::new(field.name(), dt.clone())
+            }
+            schema = Cow::Owned(s);
+        }
+
+        let null_values = null_values.map(|nv| nv.process(&schema)).transpose()?;
+
+        if let Some(cols) = columns {
+            let mut prj = Vec::with_capacity(cols.len());
+            for col in cols {
+                let i = schema.index_of(&col)?;
+                prj.push(i);
+            }
+            projection = Some(prj);
+        }
+
+        Ok(CoreReader {
+            reader_bytes: Some(reader_bytes),
+            schema,
+            projection,
+            line_number: if has_header { 1 } else { 0 },
+            ignore_parser_errors,
+            skip_rows,
+            n_rows,
+            encoding,
+            n_threads,
+            has_header,
+            delimiter,
+            sample_size,
+            chunk_size,
+            low_memory,
+            comment_char,
+            quote_char,
+            null_values,
+            predicate,
+            aggregate,
+        })
+    }
+
     fn find_starting_point<'b>(&self, mut bytes: &'b [u8]) -> Result<&'b [u8]> {
         // Skip all leading white space and the occasional utf8-bom
         bytes = skip_line_ending(skip_whitespace(skip_bom(bytes)).0).0;
@@ -166,6 +276,7 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
                         &bytes[n_bytes..],
                         self.schema.fields().len(),
                         self.delimiter,
+                        self.quote_char,
                     ) {
                         bytes = &bytes[..n_bytes + pos]
                     }
@@ -221,8 +332,13 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
 
         // split the file by the nearest new line characters such that every thread processes
         // approximately the same number of rows.
-        let file_chunks =
-            get_file_chunks(bytes, n_threads, self.schema.fields().len(), self.delimiter);
+        let file_chunks = get_file_chunks(
+            bytes,
+            n_threads,
+            self.schema.fields().len(),
+            self.delimiter,
+            self.quote_char,
+        );
 
         // If the number of threads given by the user is lower than our global thread pool we create
         // new one.
@@ -258,10 +374,15 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
                     0,
                     &self.schema,
                     &str_capacities,
-                    self.delimiter,
+                    self.quote_char,
+                    self.encoding,
+                    self.ignore_parser_errors,
                 )?;
                 let df = DataFrame::new_no_checks(
-                    buffers.into_iter().map(|buf| buf.into_series()).collect(),
+                    buffers
+                        .into_iter()
+                        .map(|buf| buf.into_series())
+                        .collect::<Result<_>>()?,
                 );
                 return Ok(df);
             }
@@ -278,8 +399,9 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
                         let mut read = bytes_offset_thread;
                         let mut df: Option<DataFrame> = None;
 
+                        let mut last_read = usize::MAX;
                         loop {
-                            if read >= stop_at_nbytes {
+                            if read >= stop_at_nbytes || read == last_read {
                                 break;
                             }
 
@@ -288,26 +410,32 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
                                 chunk_size,
                                 &schema,
                                 &str_capacities,
-                                self.delimiter,
+                                self.quote_char,
+                                self.encoding,
+                                self.ignore_parser_errors,
                             )?;
 
                             let local_bytes = &bytes[read..stop_at_nbytes];
 
-                            read = parse_lines(
+                            last_read = read;
+                            read += parse_lines(
                                 local_bytes,
                                 read,
                                 delimiter,
                                 self.comment_char,
+                                self.quote_char,
                                 self.null_values.as_ref(),
                                 projection,
                                 &mut buffers,
                                 ignore_parser_errors,
-                                self.encoding,
                                 chunk_size,
                             )?;
 
                             let mut local_df = DataFrame::new_no_checks(
-                                buffers.into_iter().map(|buf| buf.into_series()).collect(),
+                                buffers
+                                    .into_iter()
+                                    .map(|buf| buf.into_series())
+                                    .collect::<Result<_>>()?,
                             );
                             if let Some(predicate) = predicate {
                                 let s = predicate.evaluate(&local_df)?;
@@ -389,32 +517,39 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
                             capacity,
                             &schema,
                             &str_capacities,
-                            self.delimiter,
+                            self.quote_char,
+                            self.encoding,
+                            self.ignore_parser_errors,
                         )?;
 
+                        let mut last_read = usize::MAX;
                         loop {
-                            if read >= stop_at_nbytes {
+                            if read >= stop_at_nbytes || read == last_read {
                                 break;
                             }
                             let local_bytes = &bytes[read..stop_at_nbytes];
 
-                            read = parse_lines(
+                            last_read = read;
+                            read += parse_lines(
                                 local_bytes,
                                 read,
                                 delimiter,
                                 self.comment_char,
+                                self.quote_char,
                                 self.null_values.as_ref(),
                                 projection,
                                 &mut buffers,
                                 ignore_parser_errors,
-                                self.encoding,
                                 // chunk size doesn't really matter anymore,
                                 // less calls if we increase the size
-                                chunk_size * 320000,
+                                usize::MAX,
                             )?;
                         }
                         Ok(DataFrame::new_no_checks(
-                            buffers.into_iter().map(|buf| buf.into_series()).collect(),
+                            buffers
+                                .into_iter()
+                                .map(|buf| buf.into_series())
+                                .collect::<Result<_>>()?,
                         ))
                     })
                     .collect::<Result<Vec<_>>>()
@@ -423,67 +558,15 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
         }
     }
 
-    #[cfg(feature = "decompress")]
-    fn decompress_and_parse(
-        &mut self,
-        predicate: Option<Arc<dyn PhysicalIoExpr>>,
-        n_threads: usize,
-        bytes: &[u8],
-    ) -> Result<DataFrame> {
-        if let Some(bytes) = decompress(bytes) {
-            if self.inferred_schema {
-                self.schema = Cow::Owned(bytes_to_schema(
-                    &bytes,
-                    self.delimiter,
-                    self.has_header,
-                    self.skip_rows,
-                    self.comment_char,
-                )?);
-            }
-
-            self.parse_csv(n_threads, &bytes, predicate.as_ref())
-        } else {
-            self.parse_csv(n_threads, bytes, predicate.as_ref())
-        }
-    }
-
     /// Read the csv into a DataFrame. The predicate can come from a lazy physical plan.
-    pub fn as_df(
-        &mut self,
-        predicate: Option<Arc<dyn PhysicalIoExpr>>,
-        aggregate: Option<&[ScanAggregation]>,
-    ) -> Result<DataFrame> {
+    pub fn as_df(&mut self) -> Result<DataFrame> {
+        let predicate = self.predicate.take();
+        let aggregate = self.aggregate.take();
         let n_threads = self.n_threads.unwrap_or_else(|| POOL.current_num_threads());
 
-        let mut df = match &self.path {
-            // we have a path so we can mmap
-            Some(p) => {
-                let p = resolve_homedir(p);
-                let file = std::fs::File::open(&p)?;
-                let mmap = unsafe { memmap::Mmap::map(&file)? };
-                let bytes = mmap[..].as_ref();
+        let reader_bytes = self.reader_bytes.take().unwrap();
 
-                #[cfg(feature = "decompress")]
-                {
-                    self.decompress_and_parse(predicate, n_threads, bytes)?
-                }
-
-                #[cfg(not(feature = "decompress"))]
-                self.parse_csv(n_threads, bytes, predicate.as_ref())?
-            }
-            None => {
-                let mut reader = self.reader.take().unwrap();
-                let bytes = &get_reader_bytes(&mut reader)?;
-
-                #[cfg(feature = "decompress")]
-                {
-                    self.decompress_and_parse(predicate, n_threads, bytes)?
-                }
-
-                #[cfg(not(feature = "decompress"))]
-                self.parse_csv(n_threads, bytes, predicate.as_ref())?
-            }
-        };
+        let mut df = self.parse_csv(n_threads, &reader_bytes, predicate.as_ref())?;
 
         if let Some(aggregate) = aggregate {
             let cols = aggregate
@@ -503,111 +586,3 @@ impl<'a, R: MmapBytesReader> CoreReader<'a, R> {
         Ok(df)
     }
 }
-
-#[allow(clippy::too_many_arguments)]
-pub fn build_csv_reader<'a, R: MmapBytesReader>(
-    mut reader: R,
-    n_rows: Option<usize>,
-    skip_rows: usize,
-    mut projection: Option<Vec<usize>>,
-    max_records: Option<usize>,
-    delimiter: Option<u8>,
-    has_header: bool,
-    ignore_parser_errors: bool,
-    schema: Option<&'a Schema>,
-    columns: Option<Vec<String>>,
-    encoding: CsvEncoding,
-    n_threads: Option<usize>,
-    path: Option<PathBuf>,
-    schema_overwrite: Option<&'a Schema>,
-    sample_size: usize,
-    chunk_size: usize,
-    low_memory: bool,
-    comment_char: Option<u8>,
-    null_values: Option<NullValues>,
-) -> Result<CoreReader<'a, R>> {
-    // check if schema should be inferred
-    let delimiter = delimiter.unwrap_or(b',');
-
-    #[cfg(feature = "decompress")]
-    let mut inferred_schema = false;
-
-    let schema = match schema {
-        Some(schema) => Cow::Borrowed(schema),
-        None => {
-            #[cfg(feature = "decompress")]
-            {
-                // We keep track of the inferred schema bool
-                // In case the file is compressed this schema inference is wrong and has to be done
-                // again after decompression.
-                inferred_schema = true;
-
-                const N: usize = 5;
-                let mut bytes = [0u8; N];
-                reader.read_exact(&mut bytes)?;
-                // restore position
-                reader.seek(SeekFrom::Current(-(N as i64)))?;
-                if decompress(&bytes).is_some() {
-                    Cow::Owned(Schema::default())
-                } else {
-                    let (inferred_schema, _) = infer_file_schema(
-                        &mut reader,
-                        delimiter,
-                        max_records,
-                        has_header,
-                        schema_overwrite,
-                        skip_rows,
-                        comment_char,
-                    )?;
-                    Cow::Owned(inferred_schema)
-                }
-            }
-            #[cfg(not(feature = "decompress"))]
-            {
-                let (inferred_schema, _) = infer_file_schema(
-                    &mut reader,
-                    delimiter,
-                    max_records,
-                    has_header,
-                    schema_overwrite,
-                    skip_rows,
-                    comment_char,
-                )?;
-                Cow::Owned(inferred_schema)
-            }
-        }
-    };
-
-    let null_values = null_values.map(|nv| nv.process(&schema)).transpose()?;
-
-    if let Some(cols) = columns {
-        let mut prj = Vec::with_capacity(cols.len());
-        for col in cols {
-            let i = schema.index_of(&col)?;
-            prj.push(i);
-        }
-        projection = Some(prj);
-    }
-
-    Ok(CoreReader {
-        schema,
-        projection,
-        reader: Some(reader),
-        line_number: if has_header { 1 } else { 0 },
-        ignore_parser_errors,
-        skip_rows,
-        n_rows,
-        encoding,
-        n_threads,
-        path,
-        has_header,
-        delimiter,
-        sample_size,
-        chunk_size,
-        low_memory,
-        comment_char,
-        null_values,
-        #[cfg(feature = "decompress")]
-        inferred_schema,
-    })
-}
